/*
    saa7146 windows dumper

    Copyright (C) 2004 Andrew de quincey <adq_dvb@lidskialf.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <windows.h>
#include <stdlib.h>
#include "Winio.h"
#include "pci.h"
#include "saa7146.h"
#include "linuxstuff.h"

int main(int argc, char* argv[])
{
	DWORD tmp;
	int bus;
	int device;
	int reg;
	int lookforvendorid;
	int lookfordeviceid;
	int i2caddr;
	int i2cmethod;
	
	if (argc != 5) {
  	fprintf(stderr, "Syntax: saa7146dump <subsys vendorid> <subsys deviceid> <i2c address> <i2c method (0 or 1)>\n");
  	exit(1);
	}
	sscanf(argv[1], "%i", &lookforvendorid);
	sscanf(argv[2], "%i", &lookfordeviceid);
	sscanf(argv[3], "%i", &i2caddr);
	sscanf(argv[4], "%i", &i2cmethod);
	
	if (!InitializeWinIo()) {
		printf("failed to init winio\n");
		exit(1);
	}
	
	// scan PCI buses to find saa7146 devices
	for(bus=0; bus < 10; bus++) {
    for(device=0; device< 32; device++) {
      // check for valid device
      DWORD vendorid = pciread(bus, device, 0, 0) & 0xffff;
      DWORD deviceid = pciread(bus, device, 0, 0) >> 16;
      DWORD subsysvendorid = pciread(bus, device, 0, 0x2c) & 0xffff;
      DWORD subsysdeviceid = pciread(bus, device, 0, 0x2c) >> 16;
      if (((vendorid == 0xffff) || (vendorid == 0x0000)) &&
          ((deviceid == 0xffff) || (deviceid == 0x0000)) &&
          ((subsysvendorid == 0xffff) || (subsysvendorid == 0x0000)) &&
          ((subsysdeviceid == 0xffff) || (subsysdeviceid == 0x0000))) {
        continue; 
      }
      
      // is it the one we're looking for?
      if ((vendorid != 0x1131) || (deviceid != 0x7146) ||
          (subsysvendorid != lookforvendorid) || (subsysdeviceid != lookfordeviceid)) {
        continue;
      }
      
      // report
      printf("device found: bus:%i device:%i vendor:%04x device:%04x subsysvendor:%04x subsysdevice: %04x\n", 
             bus, device, vendorid, deviceid, subsysvendorid, subsysdeviceid);

      // enable PCI memory space if not already
      if ((pciread(bus, device, 0, 0x04) & 6) != 6) {
        pciwrite(bus, device, 0, 0x04, pciread(bus, device, 0, 0x04) | 6);
      }
             
      // dump out PCI registers
      printf("PCI registers: \n");
      for(reg=0; reg < 256; reg+=4) {
        printf("%02X: %08X\n", reg, pciread(bus, device, 0, reg)); 
      }
      
      // dump out saa7146 registers
      printf("SAA7146 registers: \n");
      DWORD basereg = pciread(bus, device, 0, 0x10) & 0xfffffff0;
      for(reg=0; reg < 0x200; reg+=4) {
        DWORD val = 0;
        if (!GetPhysLong((PBYTE) basereg+reg, &val)) {
          printf("Unable to read address %x\n", reg);
        } else {
          printf("%04X: %08X\n", reg, val); 
        }
      }
      
      // dump out requested i2c device
      if (i2caddr != -1) {
        unsigned char buf[10];
        struct i2c_msg i2cmsgs[2];
        saa7146_dev dev;
        int i;
        
        // setup the device
        memset(&dev, 0, sizeof(dev));
        dev.ptr = (u8*) basereg;
        dev.i2c_bitrate = SAA7146_I2C_BUS_BIT_RATE_240;
        dev.bus = bus;
        dev.device = device;
        dev.function = 0;
        saa7146_init(&dev);
        
        // setup i2c stuff
        i2cmsgs[0].addr = i2caddr;
        i2cmsgs[1].addr = i2caddr;
        i2cmsgs[0].flags = 0;
        i2cmsgs[1].flags = I2C_M_RD;
        i2cmsgs[0].buf = buf;
        i2cmsgs[1].buf = buf;
        i2cmsgs[0].len = 1;
        i2cmsgs[1].len = 1;
        
        printf("i2c registers of device @0x%x: \n", i2caddr);
        for(i=0; i<256; i++) {
          buf[0] = i;
          if (i2cmethod == 0) {
            if (saa7146_i2c_transfer(&dev, i2cmsgs, 2, 1) != 2) {
              printf("i2c TX of register %x failed\n", i);
              continue;
            }
          } else if (i2cmethod == 1) {
            if (saa7146_i2c_transfer(&dev, &i2cmsgs[0], 1, 1) != 1) {
              printf("i2c TX of register %x failed\n", i);
            }
            if (saa7146_i2c_transfer(&dev, &i2cmsgs[1], 1, 1) != 1) {
              printf("i2c TX of register %x failed\n", i);
            }            
          }
          printf("%02X: %02X\n", i, buf[0]);
        }
      }
    }
	}

	ShutdownWinIo();
}
